/* ======================================================================
     This is a part of "gpop" - Gaussian output post processor
             Copyright (c) 2002-2009 by A. Miyoshi, Univ. Tokyo
 - gpoplibu contains library functions and classes mostry for prepum.
                         created: Aug.  12, 2009
                     last edited: Oct.   3, 2009
 ====================================================================== */

#include "gpoplibu.h"

/*------------------------------------------------------------------------
     Buffer gas properties
------------------------------------------------------------------------*/

tableBufMol::tableBufMol() {                       //--- (constructor) ---
  bufMolec b;
  clear();
  // - here, molecular weights averaged for natural abundance are used
  b.name = "N2"; b.molW = 28.0134;  b.sigma = 3.798; b.epsK = 71.4;
   push_back(b);
  b.name = "HE"; b.molW = 4.002602; b.sigma = 2.551; b.epsK = 10.22;
   push_back(b);
  b.name = "AR"; b.molW = 39.948;   b.sigma = 3.542; b.epsK = 93.3;
   push_back(b);
}

bufMolec tableBufMol::getByName(string n) {            //--- getByName ---
  bufMolec bnull;
  string nam = n;
  strAllToUpper(nam);
  for (iterator p = begin(); p != end(); p++) {
    if (p->name == nam) { return *p; }
  }
  return bnull;
}

/*------------------------------------------------------------------------
     RRKM calculation properties
------------------------------------------------------------------------*/

channelInfo::channelInfo() { clear(); }            //--- (constructor) ---

void channelInfo::clear() {                                //--- clear ---
  mTS.clear(); mTS.isTS = true; mProd.clear(); flagTunnel = 0;
}

rrkmInfo::rrkmInfo() { clear(); }                  //--- (constructor) ---

void rrkmInfo::clear() {                                   //--- clear ---
  baseName.erase(); title.erase(); enCnvFac = 1.;
  temperat.setDefault();
  buffer = bufm.getByName("N2");
  molec = bufm.getByName("N2");
  mReac.clear(); mReac.isUM = true;
  channel.clear(); chNbr.clear();
}

int rrkmInfo::readFrom(string basfln) {                 //--- readFrom ---
  int ntkns, ntkns2, tempRead = false, endBlock;
  string umfln = basfln + ".um", slin, key, arg, slin2, key2;
  ifstream inf;
  vector<string> tkns, tkns2;
  channelInfo ci;
  vector<channelInfo>::iterator p;

  clear();
  inf.open(umfln.c_str());
  if (!inf)
    { cout << "failed to open [" << umfln << "].\n"; return ERRE; }
  while (getLineGpop(inf, slin) == NORM) {
    tkns = splitTokens(slin, " "); key = tkns[0]; ntkns = tkns.size();
    if (key == "title") {
      title = trim(slin.substr(slin.find("title") + 6));
    } else if (key == "energyUnit") {
      if (ntkns < 2)
        { cout << "No argument for [energyUnit].\n"; return ERRE; }
      arg = tkns[1];
      if (arg == "hartree") { enCnvFac = 1.; }
      else if (arg == "kJ/mol") { enCnvFac = 1. / KJMLHRT; }
      else if (arg == "cm-1") { enCnvFac = 1. / WAVNHRT; }
      else {
        cout << "Invalid argument [" << arg << "] for [energyUnit].\n";
        return ERRE;
      }
    } else if (key.substr(0, 4) == "temp") {
      if (!tempRead) {
        if (temperat.setBy(slin) == ERRE) { return ERRE; }
        tempRead = true;
      } else {
        if (temperat.addFrom(slin) == ERRE) { return ERRE; }
      }
    } else if (key == "bufferByName") {
      if (ntkns < 2)
        { cout << "No argument for [bufferByName].\n"; return ERRE; }
      arg = tkns[1];
      buffer = bufm.getByName(arg);
      if (buffer.name == "")
        { cout << "Unrecognized buffer molecule.\n"; return ERRE; }
    } else if (key == "buffer") {
      if (ntkns < 4)
        { cout << "Insufficient arguments for [buffer].\n"; return ERRE; }
      buffer.molW = atof(tkns[1].c_str());
      buffer.sigma = atof(tkns[2].c_str());
      buffer.epsK = atof(tkns[3].c_str());
    } else if (key == "collMolec") {
      if (ntkns < 3)
        { cout << "Insufficient arguments for [colMolec].\n"; return ERRE; }
      molec.sigma = atof(tkns[1].c_str());
      molec.epsK = atof(tkns[2].c_str());
    } else if (key == "reactant{") {
      if (mReac.readFrom(inf) == ERRE) { return ERRE; }
    } else if (key == "channel{") {
      endBlock = false;
      ci.clear();
      do {
        if (getLineGpop(inf, slin2) == ERRE) { return ERRE; }
        tkns2 = splitTokens(slin2, " "); key2 = tkns2[0];
        ntkns2 = tkns2.size();
        if (key2 == "}") {
          endBlock = true;
        } else if (key2 == "transitionState{") {
          if (ci.mTS.readFrom(inf) == ERRE) { return ERRE; }
        } else if (key2 == "products{") {
          if (ci.mProd.readFrom(inf) == ERRE) { return ERRE; }
        } else {
          cout << "Invalid keyword [" << key2 << "].\n"; return ERRE;
        }
      } while (!endBlock);
      channel.push_back(ci);
    } else {
      cout << "Invalid keyword [" << key << "].\n"; return ERRE;
    }
  }
  inf.close();
  if ((mReac.status & Energy) != 0)
    { mReac.energy *= enCnvFac; }
  for (p = channel.begin(); p != channel.end(); p++) {
    if ((p->mTS.status & Energy) != 0) { p->mTS.energy *= enCnvFac; }
    if ((p->mProd.status & Energy) != 0) { p->mProd.energy *= enCnvFac; }
  }
  baseName = basfln;
  return NORM;
}

int rrkmInfo::readProp() {                              //--- readProp ---
  int ich = 0;
  double energyTS;
  vector<channelInfo>::iterator p;
  map<double, int> orderCh;
  map<double, int>::iterator po;

  orderCh.clear();
  if (mReac.readProp("gpo") == ERRE) { return ERRE; }
  if ((mReac.status & File) == 0)                 // no reactant
    { cout << "No reactant file specified.\n"; return ERRE; }
  molec.molW = mReac.molp[0].molW;
  for (p = channel.begin(); p != channel.end(); p++) {
    if (p->mTS.readProp("gpo") == ERRE) { return ERRE; }
    if ((p->mTS.status & File) == 0) {            // no TS
      cout << "No TS file specified for channel-" << ich << ".\n";
      return ERRE;
    }
    if (p->mProd.readProp("gpo") == ERRE) { return ERRE; }
    if ((p->mProd.status & Energy) != 0)          // product energy
      { p->flagTunnel = 1; }
    energyTS = p->mTS.energy;
    orderCh.insert(make_pair(energyTS, ich));
    ich++;
  }
  chNbr.clear();
  for (po = orderCh.begin(); po != orderCh.end(); po++)
    { chNbr.push_back(po->second); }
  return NORM;
}

int rrkmInfo::outRRKM(string basfln) {                   //--- outRRKM ---
  int nTemp = temperat.size(), iTemp, nChan = chNbr.size(), idx, ich;
  int prec, nmim, imim, niss, iiss;
  double e0, V0;
  string rrkmfln;
  ofstream rof;
  molInfoSet *pmReac;
  vector<molInfoSet *> pmTS;
  vector<molInfoSet *> pmProd;
  vector<int *> pflagTunnel;
  molPropTST *p0Reac;
  vector<molPropTST *> p0TS;
  molIntMode *pmim;

  pmTS.clear(); pmProd.clear(); pflagTunnel.clear(); p0TS.clear();
  pmReac = &mReac; p0Reac = &(pmReac->molp[0]);
  for (idx = 0; idx < nChan; idx++) {
    pmTS.push_back(&channel[chNbr[idx]].mTS);
    pmProd.push_back(&channel[chNbr[idx]].mProd);
    pflagTunnel.push_back(&channel[chNbr[idx]].flagTunnel);
  }
  for (ich = 0; ich < nChan; ich++) {
    p0TS.push_back(&(pmTS[ich]->molp[0]));
  }
  rrkmfln = basfln + "_rrkm.inp";
  rof.open(rrkmfln.c_str());
  if (!rof)
    { cout << "failed to open [" << rrkmfln << "].\n"; return ERRE; }
  prec = rof.precision(); rof.setf(ios::fixed);
  rof << title << " (created by " << GPOPREV << ")\n";
  rof << " 700 100  7 " << nTemp << "  " << nChan << "  0\n";
  for (ich = 0; ich < nChan; ich++) { rof << " " << p0TS[ich]->numVib(); }
  rof << " " << p0Reac->numVib() << endl;
  e0 = pmReac->energy;
  rof.precision(3);
  for (ich = 0; ich < nChan; ich++)
    { rof << " " << (pmTS[ich]->energy - e0) * KCAMHRT; }
  rof << endl;
  for (ich = 0; ich < nChan; ich++) { rof << " " << p0TS[ich]->rotSymNbr; }
  rof << " " << p0Reac->rotSymNbr << endl;
  rof.precision(7);
  for (ich = 0; ich < nChan; ich++) {
    rof << " " << -sqrt(p0TS[ich]->rotC[1] * p0TS[ich]->rotC[2]);
  }
  rof << " " << -sqrt(p0Reac->rotC[1] * p0Reac->rotC[2]) << endl;
  for (ich = 0; ich < nChan; ich++)
    { rof << " " << p0TS[ich]->numIntRot() + 1; }
  rof << " " << p0Reac->numIntRot() + 1 << endl;
  for (ich = 0; ich < nChan; ich++) {
    rof.precision(7);
    rof << " " << p0TS[ich]->rotC[0] << " 1 1 0 0\n";
    nmim = p0TS[ich]->intMode.size();
    for (imim = 0; imim < nmim; imim++) {
      pmim = &(p0TS[ich]->intMode[imim]);
      if (pmim->type == molIntMode::FreeRotor) {
        rof.precision(7);
        rof << " " << pmim->redB << " " << pmim->nSym << " 1";
        rof.precision(0);
        if (pmim->forceV0) { rof << " " << pmim->freq << " 0\n"; }
        else { rof << " 0 0\n"; }
      } else if (pmim->type == molIntMode::HindRotor) {
        rof.precision(7);
        rof << " " << pmim->redB << " " << pmim->nSym << " 1";
        rof.precision(0);
        rof << " " << pmim->freq;
        if (pmim->forceV0) { rof << " " << pmim->V0 << endl; }
        else {
          rof << " -1\n";
        }
      }
    }
  }
  rof.precision(7);
  rof << " " << p0Reac->rotC[0] << " 1 1 0 0\n";
  nmim = p0Reac->intMode.size();
  for (imim = 0; imim < nmim; imim++) {
    pmim = &(p0Reac->intMode[imim]);
    if (pmim->type == molIntMode::FreeRotor) {
      rof.precision(7);
      rof << " " << pmim->redB << " " << pmim->nSym << " 1";
      rof.precision(0);
      if (pmim->forceV0) { rof << " " << pmim->freq << " 0\n"; }
      else { rof << " 0 0\n"; }
    } else if (pmim->type == molIntMode::HindRotor) {
      rof.precision(7);
      rof << " " << pmim->redB << " " << pmim->nSym << " 1";
      rof.precision(0);
      rof << " " << pmim->freq;
      if (pmim->forceV0) { rof << " " << pmim->V0 << endl; }
      else {
        rof << " -1\n";
      }
    }
  }
  rof.precision(3);
  rof << " " << (buffer.sigma + molec.sigma) / 2.;
  rof.precision(6);
  rof << " " << molec.molW << " " << buffer.molW;
  rof.precision(1);
  rof << " " << sqrt(buffer.epsK * molec.epsK) << endl;
  rof.precision(0);
  for (ich = 0; ich < nChan; ich++) {
    nmim = p0TS[ich]->intMode.size();
    for (imim = 0; imim < nmim; imim++) {
      pmim = &(p0TS[ich]->intMode[imim]);
      if (pmim->type == molIntMode::Vibrator)
        { rof << " " << pmim->freq << " 1\n"; }
    }
  }
  nmim = p0Reac->intMode.size();
  for (imim = 0; imim < nmim; imim++) {
    pmim = &(p0Reac->intMode[imim]);
    if (pmim->type == molIntMode::Vibrator)
      { rof << " " << pmim->freq << " 1\n"; }
  }
  rof << " 76. 760. 7.6E+3 7.6E+4 7.6E+8 7.6E+11 7.6E+14\n";
  rof.precision(3);
  for (iTemp = nTemp - 1; iTemp >= 0; iTemp--)
    { rof << " " << temperat[iTemp]; }
  rof << endl;
  for (ich = 0; ich < nChan; ich++) { rof << " " << *pflagTunnel[ich]; }
  rof << endl;
  for (ich = 0; ich < nChan; ich++) {
    if (*pflagTunnel[ich] > 0) {
      rof.precision(3);
      rof << " " << (pmProd[ich]->energy - e0) * KCAMHRT;
      rof.precision(0);
      rof << " " << pmTS[ich]->imagFreqTS << " 1\n";
    }
  }
  for (ich = 0; ich < nChan; ich++)
    { rof << " " << p0TS[ich]->isomStates.size(); }
  rof << " " << p0Reac->isomStates.size() << endl;
  for (ich = 0; ich < nChan; ich++) {
    niss = p0TS[ich]->isomStates.size();
    for (iiss = 0; iiss < niss; iiss++) {
      rof << "  " << p0TS[ich]->isomStates[iiss].deg
          << " " << p0TS[ich]->isomStates[iiss].energy;
    }
    rof << endl;
  }
  niss = p0Reac->isomStates.size();
  for (iiss = 0; iiss < niss; iiss++) {
    rof << "  " << p0Reac->isomStates[iiss].deg
          << " " << p0Reac->isomStates[iiss].energy;
  }
  rof << endl;
  rof.unsetf(ios::fixed); rof.precision(prec);
  rof.close();
  return NORM;
}

/*------------------------------------------------------------------------
     Declaration of globals
------------------------------------------------------------------------*/

tableBufMol bufm;

